/*
 * Copyright (c) 2013 CNRS/LAAS
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <ctype.h>
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include <libusb.h>

#include <daqflex/daqflex.h>

#include "daqflexInt.h"


/**
 * daqflexUSB201Init - initialise an USB-201 device
 *
 * lowChannel : lowest channel number to acquire (0..7)
 * highChannel : higest channel number to acquiree (0..7)
 * rate : acquisition rate (Hz - 0..100000)
 *
 * XXX rates above 5000 Hz won't work - performance issues ?
 */
int
daqflexUSB201Init(struct mccdevice *dev, int lowChannel,
    int highChannel, double rate)
{
	char message[64], response[64];
	int numChans = highChannel - lowChannel + 1;
	int i;

	if (lowChannel < 0 || lowChannel > highChannel)
		return -1;
	if (highChannel < lowChannel
	    || highChannel > DAQFLEX_USB201_MAX_NUM_CHANS)
		return -1;

	dev->lowChannel = lowChannel;
	dev->highChannel = highChannel;
	dev->numChans = numChans;
	dev->resolution = 0x0fff; /* 12 bits */

	if (dev->buffer != NULL)
		free(dev->buffer);
	dev->bufSize = 64;	  /* 2 * 32 words */
	dev->buffer = (int16_t *)malloc(dev->bufSize*sizeof(int16_t));
	if (dev->buffer == NULL)
		return -1;

	dev->readPtr = 0;
	dev->writePtr = 0;
	dev->numValues = 0;

	/* Use block io for higer rates */
	if (rate < 2000)
		daqflexSendMessage(dev, "AISCAN:XFRMODE=SINGLEIO",
		    response, sizeof(response));
	else
		daqflexSendMessage(dev, "AISCAN:XFRMODE=BLOCKIO",
		    response, sizeof(response));

	snprintf(message, sizeof(message), "AISCAN:LOWCHAN=%d", lowChannel);
	daqflexSendMessage(dev, message, response, sizeof(response));

	snprintf(message, sizeof(message), "AISCAN:HIGHCHAN=%d", highChannel);
	daqflexSendMessage(dev, message, response, sizeof(response));

	snprintf(message, sizeof(message), "AISCAN:RATE=%.2f", rate);
	daqflexSendMessage(dev, message, response, sizeof(response));

	/* read calibration data */
	for (i = lowChannel; i <= highChannel; i++) {
		snprintf(message, sizeof(message), "?AI{%d}:SLOPE", i);
		daqflexSendMessage(dev, message, response, sizeof(response));
		sscanf(response+12, "%lf", &dev->slope[i]);
		snprintf(message, sizeof(message), "?AI{%d}:OFFSET", i);
		daqflexSendMessage(dev, message, response, sizeof(response));
		sscanf(response+13, "%lf", &dev->offset[i]);
	}
	return 0;
}

/*
 * Convert a raw value to volts
 */
static double
decodeValue(struct mccdevice *dev, int idx, int16_t value)
{
	double calibratedData, scaledAndCalibratedData;
	int channel = dev->lowChannel + idx;

	calibratedData = (double)value*dev->slope[channel]
	    + dev->offset[channel];
	scaledAndCalibratedData = (calibratedData/(double)dev->resolution)
	    *DAQFLEX_USB201_FULLSCALE + DAQFLEX_USB201_MIN_VOLTAGE;
	return scaledAndCalibratedData;
}


/**
 * daqflexReadData - read values from an USB-201 board
 *
 * dev - device descriptor
 * values - buffer of shorts (int16_t) to hold read values
 * numSamples - number of values to read; must match the number of active
 *              channels
 *
 * if needed gets more bulk data from the device and converts them to
 * int16.
 */
int
daqflexUSB201ReadData(struct mccdevice *dev, double *values, int numSamples)
{
	int i, nread;
	unsigned char data[MAX_MESSAGE_LENGTH];

	if (numSamples != dev->numChans)
		return -1;

	while (dev->numValues < numSamples) {
		/* not enough data - acquire some more */
		nread = daqflexReadPacket(dev, data, sizeof(data));
		if (nread < 0)
			return -1;
		for (i = 0; i < nread; i+=2) {
			/* store LSB value */
			dev->buffer[dev->writePtr++] = data[i] + (data[i+1]<<8);
			if (dev->writePtr == dev->bufSize)
				dev->writePtr = 0;
			dev->numValues++;
		}
	}
	/* now we have enough data - copy them out */
	for (i = 0; i < numSamples; i++) {
		values[i] = decodeValue(dev, i, dev->buffer[dev->readPtr++]);
		if (dev->readPtr == dev->bufSize)
			dev->readPtr = 0;
		dev->numValues--;
	}
	return numSamples;
}
